from pathlib import Path
from typing import List, Optional

from clickhouse_driver import Client

from clickhouse_migrations.defaults import DB_HOST, DB_PASSWORD, DB_PORT, DB_USER
from clickhouse_migrations.migration import Migration, MigrationStorage
from clickhouse_migrations.migrator import Migrator


class ClickhouseCluster:
    def __init__(
        self,
        db_host: str = DB_HOST,
        db_user: str = DB_USER,
        db_password: str = DB_PASSWORD,
        db_port: str = DB_PORT,
        db_url: Optional[str] = None,
        db_name: Optional[str] = None,
        **kwargs,
    ):
        self.db_url: Optional[str] = db_url
        self.default_db_name: Optional[str] = db_name

        if db_url:
            parts = self.db_url.split("/")
            if len(parts) == 4:
                self.default_db_name = parts[-1]
                parts = parts[0:-1]

            self.db_url = "/".join(parts)
        else:
            self.db_host = db_host
            self.db_port = db_port
            self.db_user = db_user
            self.db_password = db_password
            self.connection_kwargs = kwargs

    def connection(self, db_name: Optional[str] = None) -> Client:
        db_name = db_name if db_name is not None else self.default_db_name

        if self.db_url:
            db_url = self.db_url
            if db_name:
                db_url = db_url + "/" + db_name
            ch_client = Client.from_url(db_url)
        else:
            ch_client = Client(
                self.db_host,
                port=self.db_port,
                user=self.db_user,
                password=self.db_password,
                database=db_name,
                **self.connection_kwargs,
            )
        return ch_client

    def create_db(
        self, db_name: Optional[str] = None, cluster_name: Optional[str] = None
    ):
        db_name = db_name if db_name is not None else self.default_db_name

        with self.connection("") as conn:
            if cluster_name is None:
                conn.execute(f'CREATE DATABASE IF NOT EXISTS "{db_name}"')
            else:
                conn.execute(
                    f'CREATE DATABASE IF NOT EXISTS "{db_name}" ON CLUSTER "{cluster_name}"'
                )

    def init_schema(
        self, db_name: Optional[str] = None, cluster_name: Optional[str] = None
    ):
        db_name = db_name if db_name is not None else self.default_db_name

        with self.connection(db_name) as conn:
            migrator = Migrator(conn)
            migrator.init_schema(cluster_name)

    def show_tables(self, db_name):
        db_name = db_name if db_name is not None else self.default_db_name

        with self.connection(db_name) as conn:
            result = conn.execute("show tables")
            return [t[0] for t in result]

    def migrate(
        self,
        db_name: Optional[str],
        migration_path: Path,
        cluster_name: Optional[str] = None,
        create_db_if_no_exists: bool = True,
        multi_statement: bool = True,
        dryrun: bool = False,
    ):
        db_name = db_name if db_name is not None else self.default_db_name

        storage = MigrationStorage(migration_path)
        migrations = storage.migrations()

        return self.apply_migrations(
            db_name,
            migrations,
            cluster_name=cluster_name,
            create_db_if_no_exists=create_db_if_no_exists,
            multi_statement=multi_statement,
            dryrun=dryrun,
        )

    def apply_migrations(
        self,
        db_name: str,
        migrations: List[Migration],
        dryrun: bool = False,
        cluster_name: Optional[str] = None,
        create_db_if_no_exists: bool = True,
        multi_statement: bool = True,
    ) -> List[Migration]:
        if create_db_if_no_exists:
            if cluster_name is None:
                self.create_db(db_name)
            else:
                self.create_db(db_name, cluster_name)

        with self.connection(db_name) as conn:
            migrator = Migrator(conn, dryrun)
            migrator.init_schema(cluster_name)
            return migrator.apply_migration(migrations, multi_statement)



import logging
from typing import Dict, List, Optional, Tuple

from clickhouse_driver import Client

from clickhouse_migrations.exceptions import MigrationException
from clickhouse_migrations.migration import Migration


class Migrator:
    def __init__(self, conn: Client, dryrun: bool = False):
        self._conn: Client = conn
        self._dryrun = dryrun

    def init_schema(self, cluster_name: Optional[str] = None):
        cluster_schema = f"""CREATE TABLE IF NOT EXISTS schema_versions ON CLUSTER "{cluster_name}" (
    version UInt32,
    md5 String,
    script String,
    created_at DateTime DEFAULT now()
) ENGINE = ReplicatedMergeTree('/clickhouse/tables/{{database}}/{{table}}', '{{replica}}')
ORDER BY tuple(created_at)"""

        single_schema = """CREATE TABLE IF NOT EXISTS schema_versions (
    version UInt32,
    md5 String,
    script String,
    created_at DateTime DEFAULT now()
) ENGINE = MergeTree ORDER BY tuple(created_at)"""

        self._execute(single_schema if cluster_name is None else cluster_schema)

    def query_applied_migrations(self) -> List[Migration]:
        query = """SELECT
            version,
            script,
            md5
        FROM schema_versions"""

        result = self._execute(query, with_column_types=True)
        column_names = [c[0] for c in result[len(result) - 1]]

        migrations_as_dict = [dict(zip(column_names, d)) for d in result[0]]

        return [Migration(**d) for d in migrations_as_dict]

    def migrations_to_apply(self, incoming: List[Migration]) -> List[Migration]:
        applied = self.query_applied_migrations()

        if not applied:
            return incoming

        if len(incoming) == 0 or len(incoming) < len(applied):
            raise MigrationException(
                "Migrations have gone missing, "
                "your code base should not truncate migrations, "
                "use migrations to correct older migrations"
            )

        # create outer join
        joined_migrations: Dict[Tuple[Migration, Migration]] = {
            m.version: [m, None] for m in incoming
        }
        for m in applied:
            if m.version in joined_migrations:
                joined_migrations[m.version][1] = m
            else:
                joined_migrations[m.version] = [None, m]

        # md5 of applied function must be equal
        for version, p in joined_migrations.items():
            left, right = p
            if left and right and left.md5 != right.md5:
                raise MigrationException(
                    "Migrations md5 is not equal, " f"Migration version is {version}."
                )

        # all migrations should be known
        for version, p in joined_migrations.items():
            left, right = p
            if not left and right:
                raise MigrationException(
                    "There is applied migrations, which is not known by current migrations list. "
                    f"Migration version is {version}."
                )

        to_apply = [
            left for left, right in joined_migrations.values() if left and not right
        ]
        return sorted(to_apply, key=lambda x: x.version)

    def apply_migration(
        self, migrations: List[Migration], multi_statement: bool
    ) -> List[Migration]:
        new_migrations = self.migrations_to_apply(migrations)

        logging.info("Total migrations to apply: %d", len(new_migrations))

        if not new_migrations:
            return []

        for migration in new_migrations:
            logging.info("Execute migration %s", migration)

            statements = self.script_to_statements(migration.script, multi_statement)

            logging.info("Migration contains %s statements to apply", len(statements))
            for statement in statements:
                if not self._dryrun:
                    self._execute(statement)
                else:
                    logging.info("Dry run mode, would have executed: %s", statement)

            logging.info("Migration applied, need to update schema version table.")
            if not self._dryrun:
                self._execute(
                    "INSERT INTO schema_versions(version, script, md5) VALUES",
                    [
                        {
                            "version": migration.version,
                            "script": migration.script,
                            "md5": migration.md5,
                        }
                    ],
                )

            logging.info("Migration is fully applied.")

        return new_migrations

    def _execute(self, statement, *args, **kwargs):
        logging.debug(statement)
        return self._conn.execute(statement, *args, **kwargs)

    @classmethod
    def script_to_statements(cls, script: str, multi_statement: bool) -> List[str]:
        statements = []
        if multi_statement:
            for statement in script.split(";"):
                statement = statement.strip()
                if statement:
                    statements.append(statement + ";")
        else:
            statements.append(script.strip())

        return statements



import logging
import os
import sys
from argparse import ArgumentParser
from pathlib import Path

from clickhouse_migrations.clickhouse_cluster import ClickhouseCluster
from clickhouse_migrations.defaults import (
    DB_HOST,
    DB_PASSWORD,
    DB_PORT,
    DB_USER,
    MIGRATIONS_DIR,
)


def log_level(value: str) -> str:
    if hasattr(logging, "getLevelNamesMapping"):
        # New api in python 3.11
        level_list = logging.getLevelNamesMapping().keys()
    else:
        level_list = logging._nameToLevel.keys()  # pylint: disable=W0212

    if value.upper() in level_list:
        return value.upper()

    raise ValueError


def cast_to_bool(value: str):
    return value.lower() in ("1", "true", "yes", "y")


def get_context(args):
    parser = ArgumentParser()
    parser.register("type", bool, cast_to_bool)

    # detect configuration
    parser.add_argument(
        "--db-url",
        default=os.environ.get("DB_URL", None),
        help="Clickhouse database hostname",
    )
    parser.add_argument(
        "--db-host",
        default=os.environ.get("DB_HOST", DB_HOST),
        help="Clickhouse database hostname",
    )
    parser.add_argument(
        "--db-port",
        default=os.environ.get("DB_PORT", DB_PORT),
        help="Clickhouse database port",
    )
    parser.add_argument(
        "--db-user",
        default=os.environ.get("DB_USER", DB_USER),
        help="Clickhouse user",
    )
    parser.add_argument(
        "--db-password",
        default=os.environ.get("DB_PASSWORD", DB_PASSWORD),
        help="Clickhouse password",
    )
    parser.add_argument(
        "--db-name",
        default=os.environ.get("DB_NAME", None),
        help="Clickhouse database name",
    )
    parser.add_argument(
        "--migrations-dir",
        default=os.environ.get("MIGRATIONS_DIR", MIGRATIONS_DIR),
        type=Path,
        help="Path to list of migration files",
    )
    parser.add_argument(
        "--multi-statement",
        default=os.environ.get("MULTI_STATEMENT", "1"),
        type=bool,
        help="Path to list of migration files",
    )
    parser.add_argument(
        "--log-level",
        default=os.environ.get("LOG_LEVEL", "WARNING"),
        type=log_level,
        help="Log level",
    )
    parser.add_argument(
        "--cluster-name",
        default=os.environ.get("CLUSTER_NAME", None),
        help="Clickhouse topology cluster",
    )
    parser.add_argument(
        "--dry-run",
        default=os.environ.get("DRY_RUN", "0"),
        type=bool,
        help="Dry run mode",
    )
    parser.add_argument(
        "--secure",
        default=os.environ.get("SECURE", "0"),
        type=bool,
        help="Secure connection",
    )

    return parser.parse_args(args)


def migrate(ctx) -> int:
    logging.basicConfig(level=ctx.log_level, style="{", format="{levelname}:{message}")

    cluster = ClickhouseCluster(
        db_host=ctx.db_host,
        db_port=ctx.db_port,
        db_user=ctx.db_user,
        db_password=ctx.db_password,
        db_url=ctx.db_url,
        secure=ctx.secure,
    )
    cluster.migrate(
        db_name=ctx.db_name,
        migration_path=ctx.migrations_dir,
        cluster_name=ctx.cluster_name,
        multi_statement=ctx.multi_statement,
        dryrun=ctx.dry_run,
    )
    return 0


def main() -> int:
    return migrate(get_context(sys.argv[1:]))  # pragma: no cover
